# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests as xdp

import pytest
import tempfile
import os
from pathlib import Path
from gi.repository import GLib, Gio

SVG_IMAGE_DATA = """<?xml version="1.0" encoding="UTF-8"?>
<svg xmlns="http://www.w3.org/2000/svg" height="16px" width="16px"/>
"""

SOUND_DATA = (
    b"\x52\x49\x46\x46\x24\x00\x00\x00\x57\x41\x56\x45" +
    b"\x66\x6d\x74\x20\x10\x00\x00\x00\x01\x00\x01\x00" +
    b"\x44\xac\x00\x00\x88\x58\x01\x00\x02\x00\x10\x00" +
    b"\x64\x61\x74\x61\x00\x00\x00\x00"
)  # fmt: skip


SUPPORTED_OPTIONS = {
    "foo": "bar",
}

NOTIFICATION_BASIC = {
    "title": GLib.Variant("s", "title"),
    "body": GLib.Variant("s", "test notification body"),
    "priority": GLib.Variant("s", "normal"),
    "default-action": GLib.Variant("s", "test-action"),
}

NOTIFICATION_BUTTONS = {
    "title": GLib.Variant("s", "test notification 2"),
    "body": GLib.Variant("s", "test notification body 2"),
    "priority": GLib.Variant("s", "low"),
    "default-action": GLib.Variant("s", "test-action"),
    "buttons": GLib.Variant(
        "aa{sv}",
        [
            {
                "label": GLib.Variant("s", "button1"),
                "action": GLib.Variant("s", "action1"),
            },
            {
                "label": GLib.Variant("s", "button2"),
                "action": GLib.Variant("s", "action2"),
            },
        ],
    ),
}


@pytest.fixture
def required_templates():
    return {
        "notification": {
            "SupportedOptions": SUPPORTED_OPTIONS,
        },
    }


class NotificationPortal(xdp.GDBusIface):
    def __init__(self):
        super().__init__(
            "org.freedesktop.portal.Desktop",
            "/org/freedesktop/portal/desktop",
            "org.freedesktop.portal.Notification",
        )

    def AddNotification(self, id, notification, fds=[]):
        return self._call(
            "AddNotification",
            GLib.Variant("(sa{sv})", (id, notification)),
            fds,
        )

    def RemoveNotification(self, id):
        return self._call(
            "RemoveNotification",
            GLib.Variant("(s)", (id,)),
        )


class TestNotification:
    def check_notification(
        self, dbus_con, app_id, id, notification_in, notification_expected
    ):
        notification_intf = NotificationPortal()
        mock_intf = xdp.get_mock_iface(dbus_con)

        method_calls = mock_intf.GetMethodCalls("AddNotification")
        backend_calls = len(method_calls)

        notification_intf.AddNotification(id, notification_in)

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("AddNotification")
        assert len(method_calls) == backend_calls + 1
        _, args = method_calls[-1]
        assert args[0] == app_id
        assert args[1] == id

        mock_notification = args[2]
        assert (
            mock_notification == GLib.Variant("a{sv}", notification_expected).unpack()
        )

    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Notification", 2)

    def test_basic(self, portals, dbus_con, app_id):
        self.check_notification(
            dbus_con,
            app_id,
            "test1",
            NOTIFICATION_BASIC,
            NOTIFICATION_BASIC,
        )

    def test_remove(self, portals, dbus_con, app_id):
        notification_intf = NotificationPortal()
        mock_intf = xdp.get_mock_iface(dbus_con)

        id = "test1"

        notification_intf.AddNotification(id, NOTIFICATION_BASIC)
        method_calls = mock_intf.GetMethodCalls("AddNotification")
        assert len(method_calls) == 1
        _, args = method_calls[-1]
        assert args[0] == app_id
        assert args[1] == id

        notification_intf.RemoveNotification(id)
        method_calls = mock_intf.GetMethodCalls("RemoveNotification")
        assert len(method_calls) == 1
        _, args = method_calls[-1]
        assert args[0] == app_id
        assert args[1] == id

    def test_buttons(self, portals, dbus_con, app_id):
        self.check_notification(
            dbus_con,
            app_id,
            "test1",
            NOTIFICATION_BUTTONS,
            NOTIFICATION_BUTTONS,
        )

    def test_markup(self, portals, dbus_con, app_id):
        bodies = [
            (
                "test <b>notification</b> body <i>italic</i>",
                "test <b>notification</b> body <i>italic</i>",
            ),
            (
                'test <a href="https://example.com"><b>Some link</b></a>',
                'test <a href="https://example.com"><b>Some link</b></a>',
            ),
            (
                "&lt;html&gt;",
                "&lt;html&gt;",
            ),
            (
                '<a href="https://xkcd.com/327/#&quot;&gt;&lt;html&gt;"></a>',
                '<a href="https://xkcd.com/327/#&quot;&gt;&lt;html&gt;"></a>',
            ),
            (
                "test \n newline \n\n some more space \n  with trailing space ",
                "test newline some more space with trailing space",
            ),
            (
                "test <custom> tag </custom>",
                "test tag",
            ),
            (
                "test <b>notification<b> body",
                False,
            ),
            (
                "<b>foo<i>bar</b></i>",
                False,
            ),
            (
                "test <markup><i>notification</i><markup> body",
                False,
            ),
        ]

        i = 0
        for body_in, body_expected in bodies:
            notification_in = NOTIFICATION_BASIC.copy()
            notification_in["markup-body"] = GLib.Variant("s", body_in)

            notification_expected = NOTIFICATION_BASIC.copy()
            if body_expected:
                notification_expected["markup-body"] = GLib.Variant("s", body_expected)

            try:
                self.check_notification(
                    dbus_con,
                    app_id,
                    f"test{i}",
                    notification_in,
                    notification_expected,
                )
                assert body_expected
            except GLib.GError as e:
                assert "invalid markup-body" in e.message

            i += 1

    def test_bad_arg(self, portals, dbus_con, app_id):
        notification = NOTIFICATION_BASIC.copy()
        notification["bodx"] = GLib.Variant("s", "Xtest")

        self.check_notification(
            dbus_con,
            app_id,
            "test1",
            notification,
            NOTIFICATION_BASIC,
        )

    def test_bad_priority(self, portals, dbus_con, app_id):
        notification = NOTIFICATION_BASIC.copy()
        notification["priority"] = GLib.Variant("s", "invalid")

        try:
            self.check_notification(
                dbus_con,
                app_id,
                "test1",
                notification,
                notification,
            )
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "invalid not a priority" in e.message

    def test_bad_button(self, portals, dbus_con, app_id):
        notification = NOTIFICATION_BUTTONS.copy()
        notification["buttons"] = GLib.Variant(
            "aa{sv}",
            [
                {
                    "labex": GLib.Variant("s", "button1"),
                    "action": GLib.Variant("s", "action1"),
                },
            ],
        )

        try:
            self.check_notification(
                dbus_con,
                app_id,
                "test1",
                notification,
                notification,
            )
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "invalid button" in e.message

    def test_display_hint(self, portals, dbus_con, app_id):
        notification = NOTIFICATION_BASIC.copy()
        notification["display-hint"] = GLib.Variant(
            "as",
            [
                "transient",
                "show-as-new",
            ],
        )

        self.check_notification(
            dbus_con,
            app_id,
            "test1",
            notification,
            notification,
        )

        notification = NOTIFICATION_BASIC.copy()
        notification["display-hint"] = GLib.Variant(
            "as",
            [
                "unsupported-hint",
            ],
        )

        try:
            self.check_notification(
                dbus_con,
                app_id,
                "test1",
                notification,
                notification,
            )
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "not a display-hint" in e.message

    def test_category(self, portals, dbus_con, app_id):
        notification = NOTIFICATION_BASIC.copy()
        notification["category"] = GLib.Variant("s", "im.received")

        self.check_notification(
            dbus_con,
            app_id,
            "test1",
            notification,
            notification,
        )

        notification = NOTIFICATION_BASIC.copy()
        notification["category"] = GLib.Variant("s", "x-vendor.custom")

        self.check_notification(
            dbus_con,
            app_id,
            "test1",
            notification,
            notification,
        )

        notification = NOTIFICATION_BASIC.copy()
        notification["category"] = GLib.Variant("s", "unsupported-type")

        try:
            self.check_notification(
                dbus_con,
                app_id,
                "test1",
                notification,
                notification,
            )
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "not a supported category" in e.message

    def test_supported_options(self, portals, dbus_con, app_id):
        properties_intf = xdp.get_iface(dbus_con, "org.freedesktop.DBus.Properties")

        options = properties_intf.Get(
            "org.freedesktop.portal.Notification", "SupportedOptions"
        )

        assert options == SUPPORTED_OPTIONS

    def test_icon_themed(self, portals, dbus_con, app_id):
        notification_intf = NotificationPortal()
        icon = Gio.ThemedIcon.new("test-icon-symbolic")

        notification = NOTIFICATION_BASIC.copy()
        notification["icon"] = icon.serialize()

        notification_intf.AddNotification("test1", notification)

    def test_icon_bytes(self, portals, dbus_con, app_id):
        notification_intf = NotificationPortal()
        bytes = GLib.Bytes.new(SVG_IMAGE_DATA.encode("utf-8"))
        icon = Gio.BytesIcon.new(bytes)

        notification = NOTIFICATION_BASIC.copy()
        notification["icon"] = icon.serialize()

        notification_intf.AddNotification("test1", notification)

    def test_icon_file(self, portals, dbus_con, app_id):
        notification_intf = NotificationPortal()
        fd, file_path = tempfile.mkstemp(prefix="notification_icon_", dir=Path.home())
        os.write(fd, SVG_IMAGE_DATA.encode("utf-8"))

        file = Gio.File.new_for_path(file_path)
        icon = Gio.FileIcon.new(file)

        notification = NOTIFICATION_BASIC.copy()
        notification["icon"] = icon.serialize()

        notification = {
            "title": GLib.Variant("s", "title"),
            "icon": icon.serialize(),
        }

        notification_intf.AddNotification("test1", notification)

    def test_icon_bad(self, portals, dbus_con, app_id):
        notification_intf = NotificationPortal()

        notification = NOTIFICATION_BASIC.copy()

        bad_icons = [
            GLib.Variant("(sv)", ["themed", GLib.Variant("s", "test-icon-symbolic")]),
            GLib.Variant(
                "(sv)",
                ["bytes", GLib.Variant("as", ["test-icon-symbolic", "test-icon"])],
            ),
            GLib.Variant("(sv)", ["file-descriptor", GLib.Variant("s", "")]),
            GLib.Variant("(sv)", ["file-descriptor", GLib.Variant("h", 0)]),
        ]

        for icon in bad_icons:
            notification["icon"] = icon
            try:
                notification_intf.AddNotification("test1", notification)
                assert False, "This statement should not be reached"
            except GLib.GError as e:
                assert e.matches(Gio.io_error_quark(), Gio.IOErrorEnum.DBUS_ERROR)

    def test_sound_simple(self, portals, dbus_con, app_id):
        notification = NOTIFICATION_BASIC.copy()
        notification["sound"] = GLib.Variant("s", "default")

        self.check_notification(
            dbus_con,
            app_id,
            "test1",
            notification,
            notification,
        )

        notification = NOTIFICATION_BASIC.copy()
        notification["sound"] = GLib.Variant("s", "silent")

        self.check_notification(
            dbus_con,
            app_id,
            "test1",
            notification,
            notification,
        )

        notification = NOTIFICATION_BASIC.copy()
        notification["sound"] = GLib.Variant("s", "bad")

        try:
            self.check_notification(
                dbus_con,
                app_id,
                "test1",
                notification,
                notification,
            )
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "invalid sound: invalid option" in e.message

    def test_sound_file(self, portals, dbus_con, app_id):
        notification_intf = NotificationPortal()
        mock_intf = xdp.get_mock_iface(dbus_con)

        fd, file_path = tempfile.mkstemp(prefix="notification_sound_", dir=Path.home())
        os.write(fd, SOUND_DATA)

        file = Gio.File.new_for_path(file_path)

        notification = NOTIFICATION_BASIC.copy()
        notification["sound"] = GLib.Variant(
            "(sv)",
            (
                "file",
                GLib.Variant("s", file.get_uri()),
            ),
        )

        notification_intf.AddNotification("test1", notification)

        method_calls = mock_intf.GetMethodCalls("AddNotification")
        assert len(method_calls) == 1
        _, args = method_calls[-1]
        mock_notification = args[2]

        assert "sound" not in mock_notification

    def test_sound_fd(self, portals, dbus_con, app_id):
        notification_intf = NotificationPortal()
        mock_intf = xdp.get_mock_iface(dbus_con)

        fd = os.memfd_create("notification_sound_test", os.MFD_ALLOW_SEALING)
        os.write(fd, SOUND_DATA)

        notification = NOTIFICATION_BASIC.copy()
        notification["sound"] = GLib.Variant(
            "(sv)",
            (
                "file-descriptor",
                GLib.Variant("h", 0),
            ),
        )

        notification_intf.AddNotification("test1", notification, [fd])

        method_calls = mock_intf.GetMethodCalls("AddNotification")
        assert len(method_calls) == 1
        _, args = method_calls[-1]
        mock_notification = args[2]

        assert mock_notification["sound"][0] == "file-descriptor"
        mock_fd = mock_notification["sound"][1]
        mock_fd = mock_fd.take()

        os.lseek(fd, 0, os.SEEK_SET)
        fd_contents = os.read(mock_fd, 1000)
        assert fd_contents == SOUND_DATA

        os.close(mock_fd)
        os.close(fd)

    def test_sound_bad(self, portals, dbus_con, app_id):
        notification_intf = NotificationPortal()

        notification = NOTIFICATION_BASIC.copy()

        bad_sounds = [
            # bad type
            GLib.Variant("(sv)", ["file-descriptor", GLib.Variant("s", "")]),
            # not sending the FD for the handle
            GLib.Variant("(sv)", ["file-descriptor", GLib.Variant("h", 13)]),
        ]

        for sound in bad_sounds:
            notification["sound"] = sound
            try:
                notification_intf.AddNotification("test1", notification)
                assert False, "This statement should not be reached"
            except GLib.GError as e:
                assert e.matches(Gio.io_error_quark(), Gio.IOErrorEnum.DBUS_ERROR)
                pass



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black
# mypy: disable-error-code="misc"

from tests.templates import Response, init_logger, ImplRequest

import dbus.service
from dataclasses import dataclass


BUS_NAME = "org.freedesktop.impl.portal.Test"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.impl.portal.Screenshot"
VERSION = 2


logger = init_logger(__name__)


@dataclass
class ScreenshotParameters:
    delay: int
    response: int
    results: dict
    expect_close: bool


def load(mock, parameters={}):
    logger.debug(f"Loading parameters: {parameters}")

    assert not hasattr(mock, "screenshot_params")
    mock.screenshot_params = ScreenshotParameters(
        delay=parameters.get("delay", 200),
        response=parameters.get("response", 0),
        results=parameters.get("results", {}),
        expect_close=parameters.get("expect-close", False),
    )

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "version": dbus.UInt32(parameters.get("version", VERSION)),
            }
        ),
    )


@dbus.service.method(
    MAIN_IFACE,
    in_signature="ossa{sv}",
    out_signature="ua{sv}",
    async_callbacks=("cb_success", "cb_error"),
)
def Screenshot(self, handle, app_id, parent_window, options, cb_success, cb_error):
    logger.debug(f"Screenshot({handle}, {app_id}, {parent_window}, {options})")
    params = self.screenshot_params

    request = ImplRequest(
        self,
        BUS_NAME,
        handle,
        logger,
        cb_success,
        cb_error,
    )

    if params.expect_close:
        request.wait_for_close()
    else:
        request.respond(Response(params.response, params.results), delay=params.delay)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="ossa{sv}",
    out_signature="ua{sv}",
    async_callbacks=("cb_success", "cb_error"),
)
def PickColor(self, handle, app_id, parent_window, options, cb_success, cb_error):
    logger.debug(f"PickColor({handle}, {app_id}, {parent_window}, {options})")
    params = self.screenshot_params

    request = ImplRequest(
        self,
        BUS_NAME,
        handle,
        logger,
        cb_success,
        cb_error,
    )

    if params.expect_close:
        request.wait_for_close()
    else:
        request.respond(Response(params.response, params.results), delay=params.delay)



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests as xdp

import dbus
import pytest


FILECHOOSER_RESULTS = {
    "uris": ["file:///test.txt", "file:///example/test2.txt"],
    "choices": [("encoding", "utf8"), ("reencode", "true"), ("third", "a")],
}


@pytest.fixture
def required_templates():
    return {
        "filechooser": {
            "results": dbus.Dictionary(FILECHOOSER_RESULTS, signature="sv"),
        },
        "lockdown": {},
    }


class TestFilechooser:
    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "FileChooser", 4)

    def test_open_file_basic(self, portals, dbus_con, app_id):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test"
        accept_label = "Accept"
        multiple = True
        options = {
            "accept_label": accept_label,
            "multiple": multiple,
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        response = request.call(
            "OpenFile",
            parent_window="",
            title=title,
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["uris"] == FILECHOOSER_RESULTS["uris"]

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("OpenFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title
        assert args[4]["accept_label"] == accept_label
        assert args[4]["multiple"] == multiple

    @pytest.mark.parametrize("template_params", ({"filechooser": {"response": 1}},))
    def test_open_file_cancel(self, portals, dbus_con, app_id):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test"
        accept_label = "Accept"
        multiple = True
        options = {
            "accept_label": accept_label,
            "multiple": multiple,
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        response = request.call(
            "OpenFile",
            parent_window="",
            title=title,
            options=options,
        )

        assert response
        assert response.response == 1

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("OpenFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title
        assert args[4]["accept_label"] == accept_label
        assert args[4]["multiple"] == multiple

    @pytest.mark.parametrize(
        "template_params", ({"filechooser": {"expect-close": True}},)
    )
    def test_open_file_close(self, portals, dbus_con, app_id):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test"
        accept_label = "Accept"
        multiple = True
        options = {
            "accept_label": accept_label,
            "multiple": multiple,
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        request.schedule_close(1000)
        request.call(
            "OpenFile",
            parent_window="",
            title=title,
            options=options,
        )

        # Only true if the impl.Request was closed too
        assert request.closed

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("OpenFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title
        assert args[4]["accept_label"] == accept_label
        assert args[4]["multiple"] == multiple

    def test_open_file_filter1(self, portals, dbus_con):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        options = {
            "filters": [
                (
                    "Images",
                    [
                        (dbus.UInt32(0), "*ico"),
                        (dbus.UInt32(1), "image/png"),
                    ],
                ),
                (
                    "Text",
                    [
                        (dbus.UInt32(0), "*.txt"),
                    ],
                ),
            ],
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        response = request.call(
            "OpenFile",
            parent_window="",
            title="Test",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["uris"] == FILECHOOSER_RESULTS["uris"]

        method_calls = mock_intf.GetMethodCalls("OpenFile")
        assert len(method_calls) == 1

    def test_open_file_filter2(self, portals, dbus_con):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")

        options = {
            "filters": [
                (
                    "Text",
                    [
                        # Invalid filter type
                        (dbus.UInt32(4), "*.txt"),
                    ],
                ),
            ],
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call(
                "OpenFile",
                parent_window="",
                title="Test",
                options=options,
            )
        assert (
            excinfo.value.get_dbus_name()
            == "org.freedesktop.portal.Error.InvalidArgument"
        )

    def test_open_file_current_filter1(self, portals, dbus_con):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        options = {
            "filters": [
                (
                    "Images",
                    [
                        (dbus.UInt32(0), "*ico"),
                        (dbus.UInt32(1), "image/png"),
                    ],
                ),
                (
                    "Text",
                    [
                        (dbus.UInt32(0), "*.txt"),
                    ],
                ),
            ],
            "current_filter": (
                "Text",
                [
                    (dbus.UInt32(0), "*.txt"),
                ],
            ),
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        response = request.call(
            "OpenFile",
            parent_window="",
            title="Test",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["uris"] == FILECHOOSER_RESULTS["uris"]

        method_calls = mock_intf.GetMethodCalls("OpenFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[4]["current_filter"] == options["current_filter"]

    def test_open_file_current_filter2(self, portals, dbus_con):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        options = {
            "current_filter": (
                "Text",
                [
                    (dbus.UInt32(0), "*.txt"),
                ],
            ),
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        response = request.call(
            "OpenFile",
            parent_window="",
            title="Test",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["uris"] == FILECHOOSER_RESULTS["uris"]

        method_calls = mock_intf.GetMethodCalls("OpenFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[4]["current_filter"] == options["current_filter"]

    def test_open_file_current_filter3(self, portals, dbus_con):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")

        options = {
            "current_filter": (
                "Text",
                [
                    # Invalid filter type
                    (dbus.UInt32(6), "*.txt"),
                ],
            ),
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call(
                "OpenFile",
                parent_window="",
                title="Test",
                options=options,
            )
        assert (
            excinfo.value.get_dbus_name()
            == "org.freedesktop.portal.Error.InvalidArgument"
        )

    def test_open_file_current_filter4(self, portals, dbus_con):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")

        options = {
            "filters": [
                (
                    "Images",
                    [
                        (dbus.UInt32(0), "*ico"),
                        (dbus.UInt32(1), "image/png"),
                    ],
                ),
                (
                    "Text",
                    [
                        (dbus.UInt32(0), "*.txt"),
                    ],
                ),
            ],
            "current_filter": (
                "Something else",
                [
                    (dbus.UInt32(0), "*.sth.else"),
                ],
            ),
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call(
                "OpenFile",
                parent_window="",
                title="Test",
                options=options,
            )
        assert (
            excinfo.value.get_dbus_name()
            == "org.freedesktop.portal.Error.InvalidArgument"
        )

    def test_open_file_choices1(self, portals, dbus_con):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        options = {
            "choices": [
                (
                    "encoding",
                    "Encoding",
                    [
                        ("utf8", "Unicode"),
                        ("latin15", "Western"),
                    ],
                    "latin15",
                ),
                (
                    "reencode",
                    "Reencode",
                    [],
                    "false",
                ),
                (
                    "third",
                    "Third",
                    [("a", "A"), ("b", "B")],
                    "",
                ),
            ],
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        response = request.call(
            "OpenFile",
            parent_window="",
            title="Test",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["uris"] == FILECHOOSER_RESULTS["uris"]

        method_calls = mock_intf.GetMethodCalls("OpenFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[4]["choices"] == options["choices"]

    def test_open_file_choices_invalid(self, portals, dbus_con):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")

        invalid_choices = [
            (
                "encoding",
                "Encoding",
                [
                    ("utf8", ""),
                    ("latin15", "Western"),
                ],
                "latin15",
            ),
            (
                "encoding",
                "Encoding",
                [
                    ("", "Unicode"),
                    ("latin15", "Western"),
                ],
                "latin15",
            ),
            (
                "",
                "Encoding",
                [
                    ("utf8", "Unicode"),
                    ("latin15", "Western"),
                ],
                "latin15",
            ),
        ]

        for choice in invalid_choices:
            request = xdp.Request(dbus_con, filechooser_intf)
            with pytest.raises(dbus.exceptions.DBusException) as excinfo:
                options = {
                    "choices": [choice],
                }
                request.call(
                    "OpenFile",
                    parent_window="",
                    title="Test",
                    options=options,
                )
            assert (
                excinfo.value.get_dbus_name()
                == "org.freedesktop.portal.Error.InvalidArgument"
            )

    def test_save_file_basic(self, portals, dbus_con, app_id):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test"
        accept_label = "Accept"
        current_name = "File Name"
        options = {
            "accept_label": accept_label,
            "current_name": current_name,
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        response = request.call(
            "SaveFile",
            parent_window="",
            title=title,
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["uris"] == FILECHOOSER_RESULTS["uris"]

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("SaveFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title
        assert args[4]["accept_label"] == accept_label
        assert args[4]["current_name"] == current_name

    @pytest.mark.parametrize("template_params", ({"filechooser": {"response": 1}},))
    def test_save_file_cancel(self, portals, dbus_con, app_id):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test"
        accept_label = "Accept"
        current_name = "File Name"
        options = {
            "accept_label": accept_label,
            "current_name": current_name,
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        response = request.call(
            "SaveFile",
            parent_window="",
            title=title,
            options=options,
        )

        assert response
        assert response.response == 1

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("SaveFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title
        assert args[4]["accept_label"] == accept_label
        assert args[4]["current_name"] == current_name

    @pytest.mark.parametrize(
        "template_params", ({"filechooser": {"expect-close": True}},)
    )
    def test_save_file_close(self, portals, dbus_con, app_id):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test"
        accept_label = "Accept"
        current_name = "File Name"
        options = {
            "accept_label": accept_label,
            "current_name": current_name,
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        request.schedule_close(1000)
        request.call(
            "SaveFile",
            parent_window="",
            title=title,
            options=options,
        )

        # Only true if the impl.Request was closed too
        assert request.closed

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("SaveFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title
        assert args[4]["accept_label"] == accept_label
        assert args[4]["current_name"] == current_name

    def test_save_file_filters(self, portals, dbus_con, app_id):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        options = {
            "filters": [
                (
                    "Images",
                    [
                        (dbus.UInt32(0), "*ico"),
                        (dbus.UInt32(1), "image/png"),
                    ],
                ),
                (
                    "Text",
                    [
                        (dbus.UInt32(0), "*.txt"),
                    ],
                ),
            ],
        }
        request = xdp.Request(dbus_con, filechooser_intf)
        response = request.call(
            "SaveFile",
            parent_window="",
            title="Title",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["uris"] == FILECHOOSER_RESULTS["uris"]

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("SaveFile")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[4]["filters"] == options["filters"]

    @pytest.mark.parametrize(
        "template_params", ({"lockdown": {"disable-save-to-disk": True}},)
    )
    def test_save_file_lockdown(self, portals, dbus_con, app_id):
        filechooser_intf = xdp.get_portal_iface(dbus_con, "FileChooser")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, filechooser_intf)
        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call(
                "SaveFile",
                parent_window="",
                title="Title",
                options={},
            )
        assert (
            excinfo.value.get_dbus_name() == "org.freedesktop.portal.Error.NotAllowed"
        )

        # Check the impl portal was not called
        method_calls = mock_intf.GetMethodCalls("FileChooser")
        assert len(method_calls) == 0



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests as xdp

import dbus
import pytest
import os
from pathlib import Path
from gi.repository import GLib


@pytest.fixture
def required_templates():
    return {"background": {}}


class TestBackground:
    def get_autostart_path(self, app_id):
        return Path(os.environ["XDG_CONFIG_HOME"]) / "autostart" / f"{app_id}.desktop"

    def get_autostart_keyfile(self, app_id):
        keyfile = GLib.KeyFile.new()

        desktop_file_path = self.get_autostart_path(app_id)
        with open(str(desktop_file_path.absolute())) as desktop_file:
            desktop_file_contents = desktop_file.read()

            assert keyfile.load_from_data(
                desktop_file_contents,
                len(desktop_file_contents),
                GLib.KeyFileFlags.NONE,
            )

        return keyfile

    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Background", 2)

    def test_request_background(self, portals, dbus_con, app_id):
        background_intf = xdp.get_portal_iface(dbus_con, "Background")
        desktop_file = self.get_autostart_path(app_id)

        reason = "Testing portals"

        request = xdp.Request(dbus_con, background_intf)
        options = {
            "reason": reason,
        }
        response = request.call(
            "RequestBackground",
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["background"]
        assert not response.results["autostart"]

        assert not desktop_file.exists()

    def test_autostart_desktopfile(self, portals, dbus_con, app_id):
        background_intf = xdp.get_portal_iface(dbus_con, "Background")

        reason = "Testing portals"
        autostart = True
        commandline = ["/bin/true", "test"]
        dbus_activatable = True

        request = xdp.Request(dbus_con, background_intf)
        options = {
            "reason": reason,
            "autostart": autostart,
            "commandline": commandline,
            "dbus-activatable": dbus_activatable,
        }
        response = request.call(
            "RequestBackground",
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["background"]
        assert response.results["autostart"]

        keyfile = self.get_autostart_keyfile(app_id)
        assert keyfile.get_string("Desktop Entry", "Type") == "Application"
        assert keyfile.get_string("Desktop Entry", "Name") == app_id
        assert keyfile.get_string("Desktop Entry", "X-XDP-Autostart") == app_id
        assert keyfile.get_string("Desktop Entry", "Exec") == "/bin/true test"
        assert keyfile.get_boolean("Desktop Entry", "DBusActivatable")

    def test_autostart_disable(self, portals, dbus_con, app_id):
        background_intf = xdp.get_portal_iface(dbus_con, "Background")
        desktop_file = self.get_autostart_path(app_id)

        reason = "Testing portals"
        autostart = True

        request = xdp.Request(dbus_con, background_intf)
        options = {
            "reason": reason,
            "autostart": autostart,
        }
        response = request.call(
            "RequestBackground",
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["background"]
        assert response.results["autostart"]

        assert desktop_file.exists()

        request = xdp.Request(dbus_con, background_intf)
        options = {
            "reason": reason,
        }
        response = request.call(
            "RequestBackground",
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["background"]
        assert not response.results["autostart"]

        assert not desktop_file.exists()

    def test_long_reason(self, portals, dbus_con, app_id):
        background_intf = xdp.get_portal_iface(dbus_con, "Background")

        reason = (
            "012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789"
            + "012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789"
            + "012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789"
        )
        autostart = True
        commandline = ["/bin/true", "test"]
        dbus_activatable = True

        request = xdp.Request(dbus_con, background_intf)
        options = {
            "reason": reason,
            "autostart": autostart,
            "commandline": commandline,
            "dbus-activatable": dbus_activatable,
        }
        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call(
                "RequestBackground",
                parent_window="",
                options=options,
            )
        assert (
            excinfo.value.get_dbus_name()
            == "org.freedesktop.portal.Error.InvalidArgument"
        )



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests as xdp

import dbus
import pytest


@pytest.fixture
def required_templates():
    return {
        "access": {},
        "lockdown": {},
    }


@pytest.fixture
def app_id():
    # x-d-p currently defaults to the empty app id for the camera portal for
    # host XdpAppInfos (which the XdpAppInfoTest is). So use the empty app_id
    # for now.
    return ""


class TestCamera:
    def set_permissions(self, dbus_con, appid, permissions):
        perm_store_intf = xdp.get_permission_store_iface(dbus_con)
        perm_store_intf.SetPermission(
            "devices",
            True,
            "camera",
            appid,
            permissions,
        )

    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Camera", 1)

    def test_access(self, portals, dbus_con, app_id):
        camera_intf = xdp.get_portal_iface(dbus_con, "Camera")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, camera_intf)
        response = request.call(
            "AccessCamera",
            options={},
        )

        assert response
        assert response.response == 0

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("AccessDialog")
        assert len(method_calls) == 1
        _, args = method_calls[-1]
        assert args[1] == app_id

    @pytest.mark.parametrize("template_params", ({"access": {"response": 1}},))
    def test_access_cancel(self, portals, dbus_con, app_id):
        camera_intf = xdp.get_portal_iface(dbus_con, "Camera")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, camera_intf)
        response = request.call(
            "AccessCamera",
            options={},
        )

        assert response
        assert response.response == 1

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("AccessDialog")
        assert len(method_calls) == 1
        _, args = method_calls[-1]
        assert args[1] == app_id

    @pytest.mark.parametrize("template_params", ({"access": {"expect-close": True}},))
    def test_access_close(self, portals, dbus_con, app_id):
        camera_intf = xdp.get_portal_iface(dbus_con, "Camera")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, camera_intf)
        request.schedule_close(1000)
        request.call(
            "AccessCamera",
            options={},
        )

        # Only true if the impl.Request was closed too
        assert request.closed

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("AccessDialog")
        assert len(method_calls) == 1
        _, args = method_calls[-1]
        assert args[1] == app_id

    @pytest.mark.parametrize(
        "template_params", ({"lockdown": {"disable-camera": True}},)
    )
    def test_access_lockdown(self, portals, dbus_con, app_id):
        camera_intf = xdp.get_portal_iface(dbus_con, "Camera")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, camera_intf)
        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call(
                "AccessCamera",
                options={},
            )
        assert (
            excinfo.value.get_dbus_name() == "org.freedesktop.portal.Error.NotAllowed"
        )

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("AccessDialog")
        assert len(method_calls) == 0

    def test_access_denied(self, portals, dbus_con, app_id):
        camera_intf = xdp.get_portal_iface(dbus_con, "Camera")
        mock_intf = xdp.get_mock_iface(dbus_con)

        self.set_permissions(dbus_con, app_id, ["no"])

        request = xdp.Request(dbus_con, camera_intf)
        response = request.call(
            "AccessCamera",
            options={},
        )

        assert response
        assert response.response == 1

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("AccessDialog")
        assert len(method_calls) == 0



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests as xdp

import pytest
import dbus
from pathlib import Path
import os
from gi.repository import GLib, Gio


EXPORT_FILES_FLAG_EXPORT_DIR = 8


def path_from_null_term_bytes(bytes):
    path_bytes, rest = bytes.split(b"\x00")
    assert rest == b""
    return Path(os.fsdecode(path_bytes))


def get_mountpoint(documents_intf):
    mountpoint = documents_intf.GetMountPoint(byte_arrays=True)
    mountpoint = path_from_null_term_bytes(mountpoint)
    assert mountpoint.exists()
    return mountpoint


def write_bytes_atomic(file_path, bytes):
    GLib.file_set_contents(file_path.absolute().as_posix(), bytes)


def write_bytes_trunc(file_path, bytes):
    try:
        fd = os.open(
            file_path.absolute().as_posix(), os.O_RDWR | os.O_TRUNC | os.O_CREAT
        )
        os.write(fd, bytes)
    finally:
        os.close(fd)


def get_host_path_attr(path):
    xattr = "xattr::document-portal.host-path"
    file = Gio.file_new_for_path(path.absolute().as_posix())
    info = file.query_info(xattr, Gio.FileQueryInfoFlags.NONE)
    host_path = info.get_attribute_as_string(xattr)
    if not host_path:
        return None
    return Path(os.fsdecode(host_path))


def export_file(documents_intf, file_path, unique=False):
    assert file_path.exists()

    with open(file_path.absolute().as_posix(), "r") as file:
        doc_id = documents_intf.Add(file.fileno(), not unique, False)
        assert doc_id

    return doc_id


def export_file_named(documents_intf, folder_path, name, unique=False):
    assert folder_path.exists()

    # bytestring convention is zero terminated
    name_nt = os.fsencode(name) + b"\x00"

    try:
        fd = os.open(folder_path.absolute().as_posix(), os.O_PATH | os.O_CLOEXEC)
        doc_id = documents_intf.AddNamed(fd, name_nt, not unique, False)
        assert doc_id
    finally:
        os.close(fd)

    return doc_id


def export_files(documents_intf, file_paths, perms, flags=0, app_id=""):
    fds = []
    try:
        for file_path in file_paths:
            fds.append(
                os.open(file_path.absolute().as_posix(), os.O_PATH | os.O_CLOEXEC)
            )

        result = documents_intf.AddFull(
            fds,
            flags,
            app_id,
            perms,
            byte_arrays=True,
        )
    finally:
        for fd in fds:
            os.close(fd)

    assert result
    return result


class TestDocuments:
    def test_version(self, xdg_document_portal, dbus_con):
        documents = dbus_con.get_object(
            "org.freedesktop.portal.Documents",
            "/org/freedesktop/portal/documents",
        )

        properties_intf = dbus.Interface(
            documents,
            "org.freedesktop.DBus.Properties",
        )
        portal_version = properties_intf.Get(
            "org.freedesktop.portal.Documents",
            "version",
        )
        assert int(portal_version) == 5

    def test_mount_point(self, xdg_document_portal, dbus_con):
        documents_intf = xdp.get_document_portal_iface(dbus_con)
        get_mountpoint(documents_intf)

    def test_create_doc(self, xdg_document_portal, dbus_con):
        documents_intf = xdp.get_document_portal_iface(dbus_con)
        mountpoint = get_mountpoint(documents_intf)

        content = b"content"
        file_name = "a-file"

        file_path = Path(os.environ["TMPDIR"]) / file_name
        write_bytes_atomic(file_path, content)
        doc_id = export_file(documents_intf, file_path)

        doc_path = mountpoint / doc_id
        doc_app1_path = mountpoint / "by-app" / "com.test.App1" / doc_id
        doc_app2_path = mountpoint / "by-app" / "com.test.App2" / doc_id

        # Make sure it got exported
        assert (doc_path / file_name).read_bytes() == content

        assert not (doc_path / "another-file").exists()
        assert not (mountpoint / "anotherid" / file_name).exists()

        # Make sure it is not viewable by apps
        assert not doc_app1_path.exists()
        assert not doc_app2_path.exists()

        # Create a tmp file in same dir, ensure it works and can't be seen by other apps
        write_bytes_atomic(doc_path / "tmp1", b"tmpdata1")
        assert (doc_path / "tmp1").read_bytes() == b"tmpdata1"
        assert not (doc_app1_path / "tmp1").exists()

        # Ensure App 1 and only it can see the document and tmpfile
        documents_intf.GrantPermissions(doc_id, "com.test.App1", ["read"])
        assert (doc_app1_path / file_name).read_bytes() == content
        assert not (doc_app2_path / file_name).exists()

        # Make sure App 1 can't create a tmpfile
        assert not (doc_app1_path / "tmp2").exists()
        with pytest.raises(PermissionError):
            (doc_app1_path / "tmp2").write_bytes(b"tmpdata1")
        assert not (doc_app1_path / "tmp2").exists()
        assert not (doc_path / "tmp2").exists()

        # Update the document contents, ensure this is propagated
        content = b"content2"
        write_bytes_atomic(doc_path / file_name, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert file_path.read_bytes() == content
        assert not (doc_app2_path / file_name).exists()
        assert not (doc_app2_path / "tmp1").exists()

        # Update the document contents outside fuse fd, ensure this is propagated
        content = b"content3"
        write_bytes_atomic(file_path, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert file_path.read_bytes() == content
        assert not (doc_app2_path / file_name).exists()
        assert not (doc_app2_path / "tmp1").exists()

        # Try to update the doc from an app that can't write to it
        with pytest.raises(PermissionError):
            (doc_app1_path / file_name).write_bytes(b"content4")

        # Update the doc from an app with write access
        documents_intf.GrantPermissions(doc_id, "com.test.App1", ["write"])
        content = b"content5"
        write_bytes_atomic(doc_app1_path / file_name, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert file_path.read_bytes() == content
        assert not (doc_app2_path / file_name).exists()

        # Try to create a tmp file for an app
        assert not (doc_app1_path / "tmp3").exists()
        write_bytes_atomic(doc_app1_path / "tmp3", b"tmpdata2")
        (doc_app1_path / "tmp3").read_bytes() == b"tmpdata2"
        assert not (doc_path / "tmp3").exists()

        # Re-Create a file from a fuse document file, in various ways
        doc_id2 = export_file(documents_intf, (doc_path / file_name))
        assert doc_id2 == doc_id
        doc_id3 = export_file(documents_intf, (doc_app1_path / file_name))
        assert doc_id3 == doc_id
        doc_id4 = export_file(documents_intf, file_path)
        assert doc_id4 == doc_id

        # Ensure we can make a unique document
        doc_id5 = export_file(documents_intf, file_path, unique=True)
        assert doc_id5 != doc_id

    def test_recursive_doc(self, xdg_document_portal, dbus_con):
        documents_intf = xdp.get_document_portal_iface(dbus_con)
        mountpoint = get_mountpoint(documents_intf)

        content = b"content"
        file_name = "recursive-file"

        file_path = Path(os.environ["TMPDIR"]) / file_name
        write_bytes_atomic(file_path, content)
        doc_id = export_file(documents_intf, file_path)

        doc_path = mountpoint / doc_id
        doc_app1_path = mountpoint / "by-app" / "com.test.App1" / doc_id

        assert (doc_path / file_name).read_bytes() == content

        doc_id2 = export_file(documents_intf, doc_path / file_name)
        assert doc_id2 == doc_id

        documents_intf.GrantPermissions(doc_id, "com.test.App1", ["read"])

        doc_id3 = export_file(documents_intf, doc_app1_path / file_name)
        assert doc_id3 == doc_id

    def test_create_docs(self, xdg_document_portal, dbus_con):
        documents_intf = xdp.get_document_portal_iface(dbus_con)
        mountpoint = get_mountpoint(documents_intf)

        files = {
            "doc1": b"doc1-content",
            "doc2": b"doc2-content",
        }

        file_paths = []
        for file_name, file_content in files.items():
            file_path = Path(os.environ["TMPDIR"]) / file_name
            write_bytes_atomic(file_path, file_content)
            file_paths.append(file_path)

        doc_ids, extra = export_files(
            documents_intf, file_paths, ["read"], app_id="org.other.App"
        )

        assert extra
        out_mountpoint = path_from_null_term_bytes(extra["mountpoint"])
        assert out_mountpoint == mountpoint

        assert doc_ids
        for doc_id, (file_name, file_content) in zip(doc_ids, files.items()):
            assert (mountpoint / doc_id / file_name).read_bytes() == file_content
            assert (Path(os.environ["TMPDIR"]) / file_name).read_bytes() == file_content
            app1_path = mountpoint / "by-app" / "com.test.App1" / doc_id / file_name
            app2_path = mountpoint / "by-app" / "com.test.App2" / doc_id / file_name
            assert not app1_path.exists()
            assert not app2_path.exists()
            assert not (mountpoint / doc_id / "another-file").exists()
            assert not (mountpoint / "anotherid" / file_name).exists()

            other_app_path = (
                mountpoint / "by-app" / "org.other.App" / doc_id / file_name
            )
            assert other_app_path.read_bytes() == file_content
            with pytest.raises(PermissionError):
                other_app_path.write_bytes(b"new-content")

    def test_add_named(self, xdg_document_portal, dbus_con):
        documents_intf = xdp.get_document_portal_iface(dbus_con)
        mountpoint = get_mountpoint(documents_intf)

        content = b"content"
        file_name = "add-named-1"

        folder_path = Path(os.environ["TMPDIR"])
        doc_id = export_file_named(documents_intf, folder_path, file_name)
        assert doc_id

        doc_path = mountpoint / doc_id
        doc_app1_path = mountpoint / "by-app" / "com.test.App1" / doc_id
        doc_app2_path = mountpoint / "by-app" / "com.test.App2" / doc_id

        assert doc_path.exists()
        assert not doc_app1_path.exists()
        assert not (doc_path / file_name).exists()
        assert not (doc_app1_path / file_name).exists()

        documents_intf.GrantPermissions(doc_id, "com.test.App1", ["read", "write"])

        assert doc_path.exists()
        assert doc_app1_path.exists()
        assert not (doc_path / file_name).exists()
        assert not (doc_app1_path / file_name).exists()

        # Update truncating with no previous file
        write_bytes_trunc(doc_path / file_name, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert not (doc_app2_path / file_name).exists()

        # Update truncating with previous file
        content = b"content2"
        write_bytes_trunc(doc_path / file_name, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert not (doc_app2_path / file_name).exists()

        # Update atomic with previous file
        content = b"content3"
        write_bytes_atomic(doc_path / file_name, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert not (doc_app2_path / file_name).exists()

        # Update from host
        content = b"content4"
        write_bytes_atomic(folder_path / file_name, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert not (doc_app2_path / file_name).exists()

        # Unlink doc
        (doc_path / file_name).unlink()
        assert doc_path.exists()
        assert doc_app1_path.exists()
        assert not (doc_path / file_name).exists()
        assert not (doc_app1_path / file_name).exists()

        # Update atomic with no previous file
        content = b"content5"
        write_bytes_atomic(doc_path / file_name, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert not (doc_app2_path / file_name).exists()

        # Unlink doc on host
        (folder_path / file_name).unlink()
        assert doc_path.exists()
        assert doc_app1_path.exists()
        assert not (doc_path / file_name).exists()
        assert not (doc_app1_path / file_name).exists()

        # Update atomic with unexpected no previous file
        content = b"content6"
        write_bytes_atomic(doc_path / file_name, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert not (doc_app2_path / file_name).exists()

        # Unlink doc on host again
        (folder_path / file_name).unlink()
        assert doc_path.exists()
        assert doc_app1_path.exists()
        assert not (doc_path / file_name).exists()
        assert not (doc_app1_path / file_name).exists()

        # Update truncating with unexpected no previous file
        content = b"content7"
        write_bytes_trunc(doc_path / file_name, content)
        assert (doc_path / file_name).read_bytes() == content
        assert (doc_app1_path / file_name).read_bytes() == content
        assert not (doc_app2_path / file_name).exists()

    def test_get_host_paths(self, xdg_document_portal, dbus_con):
        documents_intf = xdp.get_document_portal_iface(dbus_con)

        content = b"content"
        file_name = "host-path"

        file_path = Path(os.environ["TMPDIR"]) / file_name
        write_bytes_atomic(file_path, content)
        doc_id = export_file(documents_intf, file_path)

        host_paths = documents_intf.GetHostPaths([doc_id], byte_arrays=True)
        assert doc_id in host_paths
        doc_host_path = path_from_null_term_bytes(host_paths[doc_id])
        assert doc_host_path == file_path

    def test_host_paths_xattr(self, xdg_document_portal, dbus_con):
        documents_intf = xdp.get_document_portal_iface(dbus_con)
        mountpoint = get_mountpoint(documents_intf)

        base_path = Path(os.environ["TMPDIR"]) / "a"
        file_path = base_path / "b" / "c"
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_bytes(b"test")

        doc_ids, extra = export_files(
            documents_intf, [base_path], ["read"], flags=EXPORT_FILES_FLAG_EXPORT_DIR
        )
        doc_id = doc_ids[0]

        host_path = get_host_path_attr(mountpoint / doc_id)
        assert not host_path

        host_path = get_host_path_attr(mountpoint / doc_id / "a")
        assert host_path == base_path

        host_path = get_host_path_attr(mountpoint / doc_id / "a" / "b")
        assert host_path == base_path / "b"

        host_path = get_host_path_attr(mountpoint / doc_id / "a" / "b" / "c")
        assert host_path == base_path / "b" / "c"


try:
    xdp.ensure_fuse_supported()
except xdp.FuseNotSupportedException as e:
    pytest.skip(f"No fuse support: {e}", allow_module_level=True)



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests as xdp

import dbus
import pytest
import os
import tempfile
from pathlib import Path
from typing import Any


PRINT_PREPARE_DATA = {
    "token": dbus.UInt32(1337),
}


@pytest.fixture
def required_templates():
    return {
        "print": {
            "prepare-results": PRINT_PREPARE_DATA,
        },
        "lockdown": {},
    }


class TestPrint:
    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Print", 3)

    def test_prepare_print_basic(self, portals, dbus_con, app_id):
        print_intf = xdp.get_portal_iface(dbus_con, "Print")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test Title"
        settings: Any = {}
        page_setup: Any = {}
        options = {
            "modal": True,
            "accept_label": "Accept",
            "supported_output_file_formats": ["pdf"],
        }

        request = xdp.Request(dbus_con, print_intf)
        response = request.call(
            "PreparePrint",
            parent_window="",
            title=title,
            settings=settings,
            page_setup=page_setup,
            options=options,
        )

        assert response
        assert response.response == 0

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("PreparePrint")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title
        assert args[4] == settings
        assert args[5] == page_setup
        assert args[6]["modal"] == options["modal"]
        assert args[6]["accept_label"] == options["accept_label"]
        assert (
            args[6]["supported_output_file_formats"]
            == options["supported_output_file_formats"]
        )

    @pytest.mark.parametrize("template_params", ({"print": {"response": 1}},))
    def test_prepare_print_cancel(self, portals, dbus_con, app_id):
        print_intf = xdp.get_portal_iface(dbus_con, "Print")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test Title"

        request = xdp.Request(dbus_con, print_intf)
        response = request.call(
            "PreparePrint",
            parent_window="",
            title=title,
            settings={},
            page_setup={},
            options={},
        )

        assert response
        assert response.response == 1

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("PreparePrint")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title

    @pytest.mark.parametrize("template_params", ({"print": {"expect-close": True}},))
    def test_prepare_print_close(self, portals, dbus_con, app_id):
        print_intf = xdp.get_portal_iface(dbus_con, "Print")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test Title"

        request = xdp.Request(dbus_con, print_intf)
        request.schedule_close(1000)
        request.call(
            "PreparePrint",
            parent_window="",
            title=title,
            settings={},
            page_setup={},
            options={},
        )

        # Only true if the impl.Request was closed too
        assert request.closed

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("PreparePrint")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title

    @pytest.mark.parametrize(
        "template_params", ({"lockdown": {"disable-printing": True}},)
    )
    def test_prepare_print_lockdown(self, portals, dbus_con, app_id):
        print_intf = xdp.get_portal_iface(dbus_con, "Print")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test Title"

        request = xdp.Request(dbus_con, print_intf)
        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call(
                "PreparePrint",
                parent_window="",
                title=title,
                settings={},
                page_setup={},
                options={},
            )
        assert (
            excinfo.value.get_dbus_name() == "org.freedesktop.portal.Error.NotAllowed"
        )

        # Check the impl portal was not called
        method_calls = mock_intf.GetMethodCalls("PreparePrint")
        assert len(method_calls) == 0

    def test_print_basic(self, portals, dbus_con, app_id):
        print_intf = xdp.get_portal_iface(dbus_con, "Print")
        mock_intf = xdp.get_mock_iface(dbus_con)

        fd, file_path = tempfile.mkstemp(prefix="print_mock_file_", dir=Path.home())
        os.write(fd, b"print_mock_file")

        title = "Test Title"
        options = {
            "modal": True,
            "token": "token",
            "supported_output_file_formats": ["svg"],
        }

        request = xdp.Request(dbus_con, print_intf)
        response = request.call(
            "Print",
            parent_window="",
            title=title,
            fd=fd,
            options=options,
        )

        assert response
        assert response.response == 0

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Print")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == title
        assert args[5]["modal"] == options["modal"]
        assert (
            args[5]["supported_output_file_formats"]
            == options["supported_output_file_formats"]
        )

        backend_fd = args[4].take()
        ino = os.stat(f"/proc/self/fd/{fd}").st_ino
        ino_backend = os.stat(f"/proc/self/fd/{backend_fd}").st_ino
        os.close(fd)
        assert ino == ino_backend

    @pytest.mark.parametrize("template_params", ({"print": {"response": 1}},))
    def test_print_cancel(self, portals, dbus_con, app_id):
        print_intf = xdp.get_portal_iface(dbus_con, "Print")
        mock_intf = xdp.get_mock_iface(dbus_con)

        fd, _ = tempfile.mkstemp(prefix="print_mock_file_", dir=Path.home())
        os.write(fd, b"print_mock_file")

        title = "Test Title"

        request = xdp.Request(dbus_con, print_intf)
        response = request.call(
            "Print",
            parent_window="",
            title=title,
            fd=fd,
            options={},
        )

        assert response
        assert response.response == 1

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Print")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id

    @pytest.mark.parametrize("template_params", ({"print": {"expect-close": True}},))
    def test_print_close(self, portals, dbus_con, app_id):
        print_intf = xdp.get_portal_iface(dbus_con, "Print")
        mock_intf = xdp.get_mock_iface(dbus_con)

        fd, _ = tempfile.mkstemp(prefix="print_mock_file_", dir=Path.home())
        os.write(fd, b"print_mock_file")

        title = "Test Title"

        request = xdp.Request(dbus_con, print_intf)
        request.schedule_close(1000)
        request.call(
            "Print",
            parent_window="",
            title=title,
            fd=fd,
            options={},
        )

        # Only true if the impl.Request was closed too
        assert request.closed

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Print")
        assert len(method_calls) == 1
        _, args = method_calls.pop()
        assert args[1] == app_id

    @pytest.mark.parametrize(
        "template_params", ({"lockdown": {"disable-printing": True}},)
    )
    def test_print_lockdown(self, portals, dbus_con, app_id):
        print_intf = xdp.get_portal_iface(dbus_con, "Print")
        mock_intf = xdp.get_mock_iface(dbus_con)

        fd, _ = tempfile.mkstemp(prefix="print_mock_file_", dir=Path.home())
        os.write(fd, b"print_mock_file")

        title = "Test Title"

        request = xdp.Request(dbus_con, print_intf)
        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call(
                "Print",
                parent_window="",
                title=title,
                fd=fd,
                options={},
            )
        assert (
            excinfo.value.get_dbus_name() == "org.freedesktop.portal.Error.NotAllowed"
        )

        # Check the impl portal was not called
        method_calls = mock_intf.GetMethodCalls("Print")
        assert len(method_calls) == 0

    def test_print_prepare_and_print(self, portals, dbus_con, app_id):
        print_intf = xdp.get_portal_iface(dbus_con, "Print")
        mock_intf = xdp.get_mock_iface(dbus_con)

        title = "Test Title"

        fd, file_path = tempfile.mkstemp(prefix="print_mock_file_", dir=Path.home())
        os.write(fd, b"print_mock_file")

        request = xdp.Request(dbus_con, print_intf)
        response = request.call(
            "PreparePrint",
            parent_window="",
            title=title,
            settings={},
            page_setup={},
            options={},
        )

        assert response
        assert response.response == 0
        assert response.results["token"] == PRINT_PREPARE_DATA["token"]

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("PreparePrint")
        assert len(method_calls) == 1

        options = {
            "token": response.results["token"],
        }

        request = xdp.Request(dbus_con, print_intf)
        response = request.call(
            "Print",
            parent_window="",
            title=title,
            fd=fd,
            options=options,
        )

        assert response
        assert response.response == 0

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Print")
        assert len(method_calls) == 1
        _, args = method_calls.pop()

        assert args[5]["token"] == PRINT_PREPARE_DATA["token"]

        backend_fd = args[4].take()
        ino = os.stat(f"/proc/self/fd/{fd}").st_ino
        ino_backend = os.stat(f"/proc/self/fd/{backend_fd}").st_ino
        os.close(fd)
        assert ino == ino_backend



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

from typing import Callable, Dict, Optional, NamedTuple
from gi.repository import GLib
import dbus
import dbusmock
import logging


def init_logger(name: str) -> logging.Logger:
    """
    Common logging setup for the impl.portal templates. Use as:

        >>> from tests.templates import init_logger
        >>> logger = init_logger(__name__)
        >>> logger.debug("foo")

    """
    logging.basicConfig(
        format="%(levelname).1s|%(name)s: %(message)s", level=logging.DEBUG
    )
    logger = logging.getLogger(f"templates.{name}")
    logger.setLevel(logging.DEBUG)
    return logger


logger = init_logger("utils")


class Response(NamedTuple):
    response: int
    results: Dict


class ImplRequest:
    """
    Implementation of an ``org.freedesktop.impl.portal.Request`` object exposed
    on the object path ``handle``.

    The dbus method implementations need to be invoked asynchronously and the
    async callbacks must be passed in ``cb_success`` and ``cb_error``.

    The request either waits until it is closed by x-d-p (``wait_for_close``) or
    responds to the request (``respond``).
    """

    def __init__(
        self,
        mock: dbusmock.DBusMockObject,
        busname: str,
        handle: str,
        logger: logging.Logger,
        cb_success: Callable,
        cb_error: Callable,
    ):
        self.mock = mock
        bus = mock.connection
        proxy = bus.get_object(busname, handle)
        self.mock_interface = dbus.Interface(proxy, dbusmock.MOCK_IFACE)
        self.handle = handle
        self.logger = logger
        self.cb_success = cb_success
        self.cb_error = cb_error

    def respond(
        self,
        response: Callable | Response,
        delay: int = 0,
        done_cb: Callable | None = None,
    ) -> None:
        def reply():
            nonlocal response
            res = None
            logger.debug(f"Request {self.handle}: trying to reply")
            if callable(response):
                try:
                    res = response()
                except Exception as e:
                    logger.critical(
                        f"Request {self.handle}: failed getting response: {e}"
                    )
                    self.cb_error(e)
                    self._unexport()
                    return
            else:
                res = response

            assert res
            logger.debug(f"Request {self.handle}: replying {res}")

            self.cb_success(res.response, res.results)
            self._unexport()

            if done_cb:
                done_cb()

        self._export()

        if delay > 0:
            logger.debug(f"Request {self.handle}: scheduling delay of {delay}ms")
            GLib.timeout_add(delay, reply)
        else:
            reply()

    def wait_for_close(
        self,
        close_callback: Callable | None = None,
    ) -> None:
        def closed():
            logger.debug(f"Request {self.handle}: closed")

            self.mock.EmitSignal(
                "org.freedesktop.impl.portal.Mock",
                "RequestClosed",
                "s",
                (self.handle,),
            )

            if close_callback:
                try:
                    close_callback()
                except Exception as e:
                    logger.critical(
                        f"Request {self.handle}: failed running close callback: {e}"
                    )
                    self.cb_error(e)
                    self._unexport()
                    return

            response = Response(2, {})
            self.cb_success(response.response, response.results)
            self._unexport()

        def cb_methodcall(name, args):
            if name == "Close":
                closed()

        self._export()

        logger.debug(f"Request {self.handle}: waiting for x-d-p to call close")
        self.mock_interface.connect_to_signal("MethodCalled", cb_methodcall)

    def _export(self):
        # In the future we can pass a class extending
        # dbusmock.mockobject.DBusMockObject as mock_class to avoid going
        # through the mock MethodCalled signal
        self.mock.AddObject(
            path=self.handle,
            interface="org.freedesktop.impl.portal.Request",
            properties={},
            methods=[
                (
                    "Close",
                    "",
                    "",
                    "",
                )
            ],
        )

    def _unexport(self):
        self.mock.RemoveObject(self.handle)

    def __str__(self):
        return f"ImplRequest {self.handle}"


class ImplSession:
    """
    Implementation of a org.freedesktop.impl.portal.Session object. Do not
    instantiate this directly, instead use ``ImplSession.export()``. Typically
    like this:

        >>> s = ImplSession.export(mock, "org.freedesktop.impl.portal.Test", "/path/foo")

    Where the test or the backend implementation relies on the Closed() method
    of the ImplSession, provide a callback to be invoked.

        >>> r.export(close_callback=my_callback)

    Note that the latter only works if the test invokes methods
    asynchronously.

    .. attribute:: closed

        Set to True if the Close() method on the Session was invoked

    .. attribute:: handle

        The session's object path

    """

    def __init__(
        self,
        mock: dbusmock.DBusMockObject,
        busname: str,
        handle: str,
        app_id: str,
    ):
        self.mock = mock  # the main mock object
        self.handle = handle
        self.app_id = app_id
        self.closed = False
        self._close_callback: Optional[Callable] = None

        self.mock_object: Optional[dbusmock.DBusMockObject] = None

        bus = mock.connection
        proxy = bus.get_object(busname, handle)
        mock_interface = dbus.Interface(proxy, dbusmock.MOCK_IFACE)

        # Register for the Close() call on the impl.Session. If it gets
        # called, use the side-channel SessionClosed signal so we can notify
        # the test that the impl.Session was actually closed by the
        # xdg-desktop-portal
        def cb_methodcall(name, args):
            if name == "Close":
                self.closed = True
                logger.debug(f"Session.Close() on {self.handle}")
                if self._close_callback:
                    self._close_callback()
                self.mock.EmitSignal(
                    "org.freedesktop.impl.portal.Mock",
                    "SessionClosed",
                    "s",
                    (self.handle,),
                )
                self._unexport()

        mock_interface.connect_to_signal("MethodCalled", cb_methodcall)

    def export(
        self,
        close_callback: Optional[Callable] = None,
    ) -> "ImplSession":
        """
        Create the session on the bus. If ``close_callback`` is not None, that
        callback will be invoked in response to the Close() method called on
        this object.
        """
        self.mock.AddObject(
            path=self.handle,
            interface="org.freedesktop.impl.portal.Session",
            properties={},
            methods=[
                (
                    "Close",
                    "",
                    "",
                    "",
                )
            ],
        )
        # This is a bit awkward. We need our session's DBusMockObject for
        # EmitSignal of impl.portal.Session.Close. This is available in
        # dbusmock.get_object() since our template runs as part of the server.
        #
        # In theory, EmitSignal should work on self.mock_interface but
        # it doesn't and I can't figure out why.
        self.mock_object = dbusmock.get_object(self.handle)
        self._close_callback = close_callback
        return self

    def _unexport(self):
        self.mock.RemoveObject(path=self.handle)

    def close(self):
        """
        Send out Closed signal and remove this session from the bus.
        """
        assert self.mock_object is not None, "Session was never exported"
        logger.debug(f"Signal Session.Closed on {self.handle}")
        self.mock_object.EmitSignal(
            interface="org.freedesktop.impl.portal.Session",
            name="Closed",
            signature="",
            sigargs=(),
        )
        self.closed = True
        self._unexport()

    def __str__(self):
        return f"ImplSession {self.handle}"
